//+-----------------------------------------------------------------------------
//  MS Bookmaker
//  Copyright (c) 1998-2000  Microsoft Corporation.  All Rights Reserved.
//
//  File:       litconvertdemo.cpp
//
//  Contents:   driver code for .lit generation from OEB source
//
//------------------------------------------------------------------------------
#include "LITConvertDemo.h"
#include <shlwapi.h>

static HINSTANCE g_hLITGen = 0;

//+-----------------------------------------------------------------------------
//
//  Function:       CheckHR
//  Description:    If the passed HRESULT indicates an error, print a message
//                  and throw an exception.  The mix of C and C++ in this code
//                  is odd, but the (MS-extended) printf() format strings are
//                  the simplest way to mix narrow and wide character output...
//  Parameters:     hr
//                  pszMsg - the message to print, with a %x to accept the
//                      HRESULT, if it is an error\
//                  pwszExtra - an additional string to format into the message
//                      (usually a filename or entity name).  If provided,
//                      pszMsg must have a %S after the %x.
//
//------------------------------------------------------------------------------

void CheckHR(HRESULT hr, const char* pszMsg, const wchar_t* pwszExtra)
{
    if (FAILED(hr))
    {
        fprintf(stderr, pszMsg, hr, pwszExtra);
        throw _com_error(hr);
    }
}

//+-----------------------------------------------------------------------------
//
//  Function:       GetWriter
//  Description:    helper routine to load LITGen and get the main object.
//                  LITGen avoids using the COM instantiation to avoid
//                  systemwide component registration and the resulting
//                  versioning and compatibility headaches; the price we pay
//                  for that is this rather messy instantiation routine.
//
//------------------------------------------------------------------------------
static ILITWriterPtr GetWriter()
{
    HRESULT hr = S_OK;
    char szLitgenFilename[MAX_PATH+1];
    IUnknownPtr pUnkWriter;
    typedef HRESULT (*WriterCreator)(IUnknown**);
    WriterCreator CreateWriter;

    if (!g_hLITGen)
    {
        // most applications should install litgen.dll alongside the executable
        // that loads it.  In this case, since this is an SDK sample, we'll
        // load from a common location.  It's safest to always LoadLibrary()
        // with the full path, though, regardless of where the DLL is.
        GetModuleFileName(0, szLitgenFilename, MAX_PATH);
        PathRemoveFileSpec(szLitgenFilename); // trim LITConvertDemo.exe
        PathRemoveFileSpec(szLitgenFilename); // trim Debug (or Release)
        PathRemoveFileSpec(szLitgenFilename); // trim LITConvertDemo
        PathRemoveFileSpec(szLitgenFilename); // trim C++Demos
        PathAppend(szLitgenFilename, "LITGenComponent\\litgen.dll");
        g_hLITGen = LoadLibrary(szLitgenFilename);
    }
    if (!g_hLITGen)
    {
        hr = HRESULT_FROM_WIN32(::GetLastError());
        CheckHR(hr, "Error 0x%x loading litgen.dll.  (This sample will only "
                    "run from its default location in the SDK tree.)\n");
    }
    CreateWriter = (WriterCreator)GetProcAddress(g_hLITGen, "CreateWriter");
    if (!CreateWriter)
    {
        hr = HRESULT_FROM_WIN32(::GetLastError());
        CheckHR(hr, "Error 0x%x getting LITGen entry point.\n");
    }
    hr = CreateWriter(&pUnkWriter);
    CheckHR(hr, "Error 0x%x creating ILITWriter.\n");

    return ILITWriterPtr(pUnkWriter);
}

int main(int argc, char* argv[])
{
    int ret = EXIT_SUCCESS;
    HRESULT         hr = S_OK;

    if (argc != 3)
    {
        fprintf(stderr, "Usage:\n\t%s <packagefile> <litfile>\n", argv[0]);
        return EXIT_FAILURE;
    }

    ILITWriterPtr pWriter;
    try
    {
        hr = CoInitialize(0);
        CheckHR(hr, "Error 0x%x initializing COM.\n");

        // resolve and widen filenames...
        char szFilenameBuf[MAX_PATH+1];
        char* dummy;
        if (!GetFullPathName(argv[1], MAX_PATH, szFilenameBuf, &dummy))
        {
            hr = HRESULT_FROM_WIN32(::GetLastError());
            CheckHR(hr, "Error 0x%x resolving package filename.\n");
        }
        _bstr_t bstrPackage = szFilenameBuf;
        const wchar_t* pwszPackage = bstrPackage;

        if (!GetFullPathName(argv[2], MAX_PATH, szFilenameBuf, &dummy))
        {
            hr = HRESULT_FROM_WIN32(::GetLastError());
            CheckHR(hr, "Error 0x%x resolving .lit filename.\n");
        }
        _bstr_t bstrLit = szFilenameBuf;
        const wchar_t* pwszLit = bstrLit;

        IUnknownPtr pUnk;

        // Create a writer
        pWriter = GetWriter();

        // Tell the writer about our callback
        ILITCallbackPtr pCallback(new ErrHandler);
        assert(pCallback != 0);
        hr = pWriter->SetCallback(pCallback);
        CheckHR(hr, "Error 0x%x setting message callback on writer.\n");

        // Create the output file
        hr = pWriter->Create(pwszLit, pwszPackage, L"DemoPublisher", 0);
        CheckHR(hr, "Error 0x%x creating output file %S.\n", pwszLit);

        // Get the package host...
        hr = pWriter->GetPackageHost(FALSE, &pUnk);
        CheckHR(hr, "Error 0x%x getting package file host.\n");
        ILITParserHostPtr pParserHost(pUnk);
        pUnk = 0;

        // Parse the package file at the package host...
        ParseXML(pwszPackage, pParserHost);

        // ...and finish the host.
        hr = pParserHost->Finish();
        CheckHR(hr, "Error 0x%x processing package file %S.\n", pwszPackage);

        // now loop through our content.  We loop through all CSS, then all
        // content, then all images; the ILITWriter dictates the order.
        // CSS first:
        while (1)
        {
            // Get the next host.
            hr = pWriter->GetNextCSSHost(&pUnk);
            CheckHR(hr, "Error 0x%x getting CSS host.\n");

            // If there is no next host, we'll get S_FALSE; continue on to
            // content.
            if (S_FALSE == hr)
                break;

            // Get the real interface and release the old one.
            ILITCSSHostPtr pCSSHost(pUnk);
            pUnk = 0;

            // Fetch the filename for this CSS file...
            BSTR bstrFilename = 0;
            hr = pCSSHost->GetFilename(&bstrFilename);
            CheckHR(hr, "Error 0x%x getting CSS filename.\n");
            _bstr_t bstrHolder(bstrFilename, FALSE); // ensure cleanup

            // Copy the CSS text...
            printf("Copying CSS file %S...\n", bstrFilename);
            CopyCSS(bstrHolder, pCSSHost);

            // ...and finish the host.
            hr = pCSSHost->Finish();
            CheckHR(hr, "Error 0x%x writing CSS file %S.\n", bstrFilename);
        }

        // The CSS is done; content:
        while (1)
        {
            // Get the next host.
            hr = pWriter->GetNextContentHost(&pUnk);
            CheckHR(hr, "Error 0x%x getting content host.\n");

            // If there is no next host, we'll get S_FALSE; continue on to
            // images.
            if (S_FALSE == hr)
                break;
            
            // Get the real interface and release the old one.
            ILITParserHostPtr pParserHost(pUnk);
            pUnk = 0;

            // Fetch the filename for this content file...
            BSTR bstrFilename = 0;
            hr = pParserHost->GetFilename(&bstrFilename);
            CheckHR(hr, "Error 0x%x getting content filename.\n");
            _bstr_t bstrHolder(bstrFilename, FALSE); // ensure cleanup

            // parse the content at the host...
            printf("Processing content file %S...\n", bstrFilename);
            ParseXML(bstrFilename, pParserHost);

            // ...and finish the host.
            hr = pParserHost->Finish();
            CheckHR(hr, "Error 0x%x processing OEB content %S.\n",
                bstrFilename);
        }

        // Nothing left but images:
        while (1)
        {
            // Get the next host.
            hr = pWriter->GetNextImageHost(&pUnk);
            CheckHR(hr, "Error 0x%x getting image host.\n");

            // If there is no next host, we'll get S_FALSE; we've written
            // everything.
            if (S_FALSE == hr)
                break;

            // Get the real interface and release the old one.
            ILITImageHostPtr pImageHost(pUnk);
            pUnk = 0;

            // Fetch the filename for this image file...
            BSTR bstrFilename = 0;
            hr = pImageHost->GetFilename(&bstrFilename);
            CheckHR(hr, "Error 0x%x getting image filename.\n");
            _bstr_t bstrHolder(bstrFilename, FALSE);

            // Copy the image stream...
            printf("Copying image file %S...\n", bstrFilename);
            CopyImage(bstrHolder, pImageHost);

            // ...and we're done (image hosts don't get Finish(), for
            // historical reasons).
        }

        // Everything's done; finalize the file.
        hr = pWriter->Finish();
        CheckHR(hr, "Error 0x%x finishing output file.\n");
    }
    catch(const _com_error& e)
    {
        fprintf(stderr, "Terminating for error 0x%x.\n", e.Error());
        // In case of failure, it's up to us to call Fail() to make sure the
        // broken file is deleted, etc. and everything is cleaned up and
        // released.
        if (pWriter)
        {
            pWriter->Fail();
        }
        ret = EXIT_FAILURE;
    }
    if (pWriter)
        pWriter.Release();
    if (g_hLITGen)
        FreeLibrary(g_hLITGen);
    CoUninitialize();
    return ret;
}
