//+-----------------------------------------------------------------------------
//  MS Bookmaker
//  Copyright (c) 1998-2000  Microsoft Corporation.  All Rights Reserved.
//
//  File:       XmlDomParser.cpp
//
//  Contents:   code for loading an XML document into the DOM, then walking it
//              and feeding content into an ILITParserHost
//
//------------------------------------------------------------------------------
#include "LITConvertDemo.h"

#import <msxml.dll> named_guids

//+-----------------------------------------------------------------------------
//
//  Function:       WalkTree
//  Description:    recursive helper for ParseXML(); recursively walks the DOM
//                  tree, pumping notifications into an ILITParserHost.
//  Parameters:     pNode - pointer to a root in the DOM tree
//                  pHost - LITGen parser host
//
//------------------------------------------------------------------------------
static void WalkTree(MSXML::IXMLDOMNodePtr pNode, ILITParserHost* pHost)
{
    HRESULT hr = S_OK;

    MSXML::DOMNodeType type = pNode->GetnodeType();
    switch(type)
    {
    case MSXML::NODE_DOCUMENT:
        {
            // The NODE_DOCUMENT node can have multiple children - e.g. the
            // DTD.  We only want the root element, which must be the only one
            // of type NODE_ELEMENT.
            MSXML::IXMLDOMNodePtr pChild = pNode->GetfirstChild();
            while (pChild != 0 && pChild->GetnodeType() != MSXML::NODE_ELEMENT)
                pChild = pChild->GetnextSibling();
            if (pChild)
                WalkTree(pChild, pHost);
        }
        break;
    case MSXML::NODE_ELEMENT:
        {
            // NODE_ELEMENT corresponds directly to a tag.
            // Get a new tag from LITGen:
            IUnknownPtr pUnk;
            pHost->NewTag(&pUnk);
            ILITTagPtr pTag(pUnk);
            pUnk = 0;

            // Get the element name and set the tag name...
            _bstr_t bstrName = pNode->GetnodeName();
            hr = pTag->SetName(bstrName, bstrName.length());
            CheckHR(hr, "Error 0x%x setting tag name %S", bstrName);

            // if we have attributes, loop through and add each one:
            MSXML::IXMLDOMNamedNodeMapPtr pAttrs = pNode->Getattributes();
            if (pAttrs)
            {
                for(MSXML::IXMLDOMNodePtr pChild = pAttrs->nextNode(); pChild;
                    pChild = pAttrs->nextNode())
                {
                    assert(pChild->GetnodeType() == MSXML::NODE_ATTRIBUTE);
                    _bstr_t bstrAttrName = pChild->GetnodeName();
                    // BUGBUG: will this conversion do the right thing, even
                    // in the presence of a DTD that results in other variant
                    // types in the attribute value?
                    _bstr_t bstrAttrValue = pChild->GetnodeValue();
                    hr = pTag->AddAttribute(
                             bstrAttrName, bstrAttrName.length(),
                             bstrAttrValue, bstrAttrValue.length());
                    CheckHR(hr, "Error 0x%x setting tag attribute %S",
                            bstrAttrName);
                }
            }
            // Send the tag back to LITGen.  If the element has no children,
            // it's more efficient to say so in the Tag() call, so we'll find
            // that out first:
            MSXML::IXMLDOMNodePtr pChild = pNode->GetfirstChild();
            pHost->Tag(pTag, pChild ? TRUE : FALSE);

            // If we have children, loop through and handle each one...
            if (pChild)
            {
                // Recurse for all the children of this tag...
                while (pChild)
                {
                    WalkTree(pChild, pHost);
                    pChild = pChild->GetnextSibling();
                }

                // After we send all children of a tag, we must call
                // EndChildren().  Note that we only do this if we sent TRUE
                // to Tag() above.
                hr = pHost->EndChildren();
                CheckHR(hr, "Error 0x%x closing tag %S", bstrName);
            }
        }
        break;
    // Each of these types is different in some irrelevant way - to us,
    // they're all just text.
    case MSXML::NODE_TEXT:
    case MSXML::NODE_CDATA_SECTION:
    case MSXML::NODE_ENTITY_REFERENCE:
        {
            // fetch all the text and send it to the host.
            // The node we're looking at may not be a leaf, but Gettext() will
            // traverse a subtree if necessary to make it appear atomic.
            _bstr_t bstrText = pNode->Gettext();
            if (bstrText.length())
            {
                hr = pHost->Text(bstrText, bstrText.length());
                CheckHR(hr, "Error 0x%x sending text %S", bstrText);
            }
        }
        break;
    }
}

//+-----------------------------------------------------------------------------
//
//  Function:       ParseXML
//  Description:    Loads an XML source document into a DOM document, then
//                  feeding its content into the parser host.
//  Parameters:     pwszURL - wide string holding the name of the file to open
//                  pHost - parser host from LITGen
//
//------------------------------------------------------------------------------
void ParseXML(const wchar_t* pwszURL, ILITParserHost* pHost)
{
    HRESULT hr = S_OK;

    // Create the XML DOM document...
    MSXML::IXMLDOMDocumentPtr pDoc;
    hr = CoCreateInstance(MSXML::CLSID_DOMDocument, NULL,
        CLSCTX_INPROC_SERVER, MSXML::IID_IXMLDOMDocument,
        reinterpret_cast<void**>(&pDoc));
    CheckHR(hr, "Error 0x%x creating XML DOM document.\n");

    // We don't have any use for asynchronous load; just block for it.
    pDoc->Putasync(VARIANT_FALSE);
    // We don't need to do XML-style validation (actually, OEB documents
    // don't even have to be valid in the XML sense).
    pDoc->PutvalidateOnParse(VARIANT_FALSE);
    pDoc->PutpreserveWhiteSpace(VARIANT_TRUE);
    if (!pDoc->load(pwszURL))
    {
        MSXML::IXMLDOMParseErrorPtr pError = pDoc->GetparseError();
        fprintf(stderr, "Parse error %d at %S:%d: %S\n",
            pError->GeterrorCode(), pwszURL, pError->Getline(),
            static_cast<const wchar_t*>(pError->Getreason()));
        throw _com_error(E_FAIL);
    }

    // we have a good load; recurse through the tree.
    WalkTree(pDoc, pHost);
}
