//+-----------------------------------------------------------------------------
//  MS Bookmaker
//  Copyright (c) 1998-2000  Microsoft Corporation.  All Rights Reserved.
//
//  File:       literrordemo.cpp
//
//  Contents:   driver code for .lit generation from synthetic arrays
//
//  This sample demonstrates driving LITGen from an entirely arbitrary data
//  source.  The notifications to be sent are specified in arrays; there's one
//  array of package "file" notifications, a CSS "file", and three different
//  arrays of content "file" notifications.  A commandline argument selects
//  which of the content arrays to use, and what output filename to use.
//
//------------------------------------------------------------------------------
#define _WIN32_DCOM
#include <windows.h>
#include <objbase.h>
#include <shlwapi.h>
#include <iostream>
#include <assert.h>
#include "litgen.h"

//+-----------------------------------------------------------------------------
//
//  Class:      Callback
//
//  Summary:    A simple implementation of ILITCallback; it simply dumps
//              whatever it receives to stdout or stderr.
//
//------------------------------------------------------------------------------
class Callback : public ILITCallback
{
    STDMETHOD(QueryInterface)(REFIID riid, void** ppv);
    STDMETHOD_(ULONG, AddRef)() { return 2; }
    STDMETHOD_(ULONG, Release)() { return 1; }
    STDMETHOD(Message)(int iType, int iMessageCode, const wchar_t* pwszText);
} g_callback;

STDMETHODIMP Callback::QueryInterface(REFIID riid, void** ppv)
{
    if (!ppv)
        return E_POINTER;
    *ppv = 0;
    if (riid == IID_ILITCallback || riid == IID_IUnknown)
        *ppv = static_cast<ILITCallback*>(this);
    if (!*ppv)
        return E_NOINTERFACE;
    // no need to AddRef() a stack object
    return S_OK;
}

STDMETHODIMP Callback::Message(int iType, int iMessageCode,
                               const wchar_t* pwszText)
{
    switch (iType)
    {
        case -1:
            std::wcout << L"LITGen Message: " << pwszText << std::endl;
            break;
        case 0:
            std::wcerr << L"LITGen Error: "   << pwszText << std::endl;
            break;
        case 1:
            std::wcout << L"LITGen Warning: " << pwszText << std::endl;
            break;
    }
    return S_OK;
}

// these terminated arrays of name/value pairs will be used in the CallInfo
// arrays below.
struct Attribute {
    const wchar_t* pwszName;
    const wchar_t* pwszValue;
}
package_attrs[] =
{
    { L"unique-identifier", L"notisbn" },
    { 0, 0 }
},
identifier_attrs[] =
{
    { L"id", L"notisbn" },
    { 0, 0 }
},
creator_attrs[] =
{
    { L"file-as", L"Microsoft Corporation" },
    { 0, 0 }
},
item_attrs1[] =
{
    { L"id", L"content" },
    { L"href", L"sample.htm" },
    { L"media-type", L"text/x-oeb1-document" },
    { 0, 0 }
},
item_attrs2[] =
{
    { L"id", L"css" },
    { L"href", L"sample.css" },
    { L"media-type", L"text/x-oeb1-css" },
    { 0, 0 }
},
itemref_attrs[] =
{
    { L"idref", L"content" },
    { 0, 0 }
},
link_attrs[] =
{
    { L"href", L"sample.css" },
    { L"rel", L"stylesheet" },
    { L"type", L"text/x-oeb1-css" },
    { 0, 0 }
},
badimg_attrs[] =
{
    { L"src", L"sample.css" },
    { 0, 0 }
};

// this rather arbitrary structure contains enough information to be used for
// simple "scripting" of LITGen notifications.  Remember that unspecified
// members are zero-filled when doing this kind of initialization.
struct CallInfo {
    enum Type { TAG=1, TEXT, ENDCHILDREN } type; // which notification?
    const wchar_t* pwszName;                     // what tag name, or what text?
    bool fChildren;                              // if tag, are there children?
    Attribute* paAttributes;                     // attribute array for tag?
}
package_calls[] =
{
    { CallInfo::TAG, L"package", true, package_attrs },
    {   CallInfo::TAG, L"metadata", true },
    {     CallInfo::TAG, L"dc-metadata", true },
    {       CallInfo::TAG, L"dc:Title", true },
    {         CallInfo::TEXT, L"Tiny LIT" },
    {       CallInfo::ENDCHILDREN, L"dc:Title" },
    {       CallInfo::TAG, L"dc:Creator", true, creator_attrs },
    {         CallInfo::TEXT, L"Microsoft Corporation" },
    {       CallInfo::ENDCHILDREN, L"dc:Creator" },
    {       CallInfo::TAG, L"dc:Identifier", true, identifier_attrs },
    {         CallInfo::TEXT, L"0-123-45678-9" },
    {       CallInfo::ENDCHILDREN, L"dc:Identifier" },
    {     CallInfo::ENDCHILDREN, L"dc-metadata" },
    {   CallInfo::ENDCHILDREN, L"metadata" },
    {   CallInfo::TAG, L"manifest", true },
    {     CallInfo::TAG, L"item", false, item_attrs1 },
    {     CallInfo::TAG, L"item", false, item_attrs2 },
    {   CallInfo::ENDCHILDREN, L"manifest" },
    {   CallInfo::TAG, L"spine", true },
    {     CallInfo::TAG, L"itemref", false, itemref_attrs },
    {   CallInfo::ENDCHILDREN, L"spine" },
    { CallInfo::ENDCHILDREN, L"package" },
    { CallInfo::Type(0), 0 }
},
content_calls_good[] =
{
    { CallInfo::TAG, L"html", true },
    {   CallInfo::TAG, L"head", true },
    {     CallInfo::TAG, L"link", false, link_attrs },
    {   CallInfo::ENDCHILDREN, L"head" },
    {   CallInfo::TAG, L"body", true },
    {     CallInfo::TEXT, L"Sample LIT content" },
    {   CallInfo::ENDCHILDREN, L"body" },
    { CallInfo::ENDCHILDREN, L"html" },
    { CallInfo::Type(0), 0 }
},
content_calls_warn[] =
{
    { CallInfo::TAG, L"html", true },
    {   CallInfo::TAG, L"body", true },
    {     CallInfo::TAG, L"broken", false }, // no such tag - generates warning
    {     CallInfo::TEXT, L"Sample LIT content" },
    {   CallInfo::ENDCHILDREN, L"body" },
    { CallInfo::ENDCHILDREN, L"html" },
    { CallInfo::Type(0), 0 }
},
content_calls_fail[] =
{
    { CallInfo::TAG, L"html", true },
    {   CallInfo::TAG, L"body", true },
    {     CallInfo::TEXT, L"Sample LIT content" },
    {     CallInfo::TAG, L"img", false, badimg_attrs },
    {   CallInfo::ENDCHILDREN, L"body" },
    { CallInfo::ENDCHILDREN, L"html" },
    { CallInfo::Type(0), 0 }
};

// these are the filenames we'll use for the output files from each array.
const wchar_t* pwszFilenameGood = L"goodarray.lit";
const wchar_t* pwszFilenameWarn = L"warnarray.lit";
const wchar_t* pwszFilenameFail = L"failarray.lit";

const unsigned char szCSS[] = "body { font-weight: bold; font-size: x-large }";
const int cbCSS = sizeof(szCSS) - 1;

//+-----------------------------------------------------------------------------
//
//  Function:       Pump
//  Description:    Sends all the notifications in a terminated array of
//                  CallInfo entries to the supplied LITGen parser host.  This
//                  works the same for both package and content streams.
//
//------------------------------------------------------------------------------
HRESULT Pump(ILITParserHost* pHost, CallInfo pCallInfo[])
{
    HRESULT hr = S_OK;
    while (S_OK == hr)
    {
        switch(pCallInfo->type)
        {
        case 0:
            return hr;
        case CallInfo::TAG:
        {
            IUnknown* pUnkTag = 0;
            ILITTag* pTag = 0;
            // Create a new tag...
            hr = pHost->NewTag(&pUnkTag);
            // query for the real interface...
            if (SUCCEEDED(hr))
                hr = pUnkTag->QueryInterface(IID_ILITTag,
                         reinterpret_cast<void**>(&pTag));
        
            // set the tag name...
            if (SUCCEEDED(hr))
                hr = pTag->SetName(pCallInfo->pwszName,
                                   wcslen(pCallInfo->pwszName));

            // loop through attributes, if supplied
            if (SUCCEEDED(hr) && pCallInfo->paAttributes)
            {
                Attribute* pAttr = pCallInfo->paAttributes;
                while (SUCCEEDED(hr) && pAttr->pwszName)
                {
                    hr = pTag->AddAttribute(pAttr->pwszName,
                                            wcslen(pAttr->pwszName),
                                            pAttr->pwszValue,
                                            wcslen(pAttr->pwszValue));
                    ++pAttr;
                }
            }

            // send the tag to the host
            if (SUCCEEDED(hr))
                hr = pHost->Tag(pTag, pCallInfo->fChildren);

            if (pTag)
                pTag->Release();
            if (pUnkTag)
                pUnkTag->Release();
            break;
        }
        case CallInfo::TEXT:
            hr = pHost->Text(pCallInfo->pwszName, wcslen(pCallInfo->pwszName));
            break;
        case CallInfo::ENDCHILDREN:
            hr = pHost->EndChildren();
            break;
        }
        ++pCallInfo;
    }
    return hr;
}

int main(int argc, char* argv[])
{
    HRESULT hr = S_OK;

    CoInitializeEx(0, COINIT_APARTMENTTHREADED);
    ILITWriter* pWriter = 0;
    HMODULE hLITGen = 0;

    CallInfo* pContentCalls = 0;
    const wchar_t* pwszFilename = 0;

    if (argc == 2 && argv[1][0] && argv[1][1] == '\0')
    {
        switch(argv[1][0])
        {
        case 'G':
        case 'g':
            pContentCalls = content_calls_good;
            pwszFilename = pwszFilenameGood;
            break;
        case 'W':
        case 'w':
            pContentCalls = content_calls_warn;
            pwszFilename = pwszFilenameWarn;
            break;
        case 'F':
        case 'f':
            pContentCalls = content_calls_fail;
            pwszFilename = pwszFilenameFail;
            break;
        }
    }
    if (!pContentCalls)
    {
        std::cerr << "Usage:\t" << argv[0] << " [G|W|F]\n";
        std::cerr << "\tG: create goodarray.lit with no warnings or errors\n";
        std::cerr << "\tW: create warnarray.lit with warnings\n";
        std::cerr << "\tF: try to create failarray.lit and fail\n";
        return -1;
    }

    // load litgen, create the writer
    {
        IUnknown* pUnkWriter = 0;
        typedef HRESULT (*WriterCreator)(IUnknown**);
        WriterCreator CreateWriter;

        // most applications should install litgen.dll alongside the executable
        // that loads it.  In this case, since this is an SDK sample, we'll
        // load from a common location.  It's safest to always LoadLibrary()
        // with the full path, though, regardless of where the DLL is.
        char szLitgenFilename[MAX_PATH+1];
        GetModuleFileName(0, szLitgenFilename, MAX_PATH);
        PathRemoveFileSpec(szLitgenFilename); // trim LITErrorDemo.exe
        PathRemoveFileSpec(szLitgenFilename); // trim Debug (or Release)
        PathRemoveFileSpec(szLitgenFilename); // trim LITErrorDemo
        PathRemoveFileSpec(szLitgenFilename); // trim C++Demos
        PathAppend(szLitgenFilename, "LITGenComponent\\litgen.dll");
        hLITGen = LoadLibrary(szLitgenFilename);
        if (hLITGen)
        {
            CreateWriter = (WriterCreator)GetProcAddress(hLITGen, "CreateWriter");
            if (CreateWriter)
                hr = CreateWriter(&pUnkWriter);
            else
                hr = HRESULT_FROM_WIN32(GetLastError());
        }
        else
            hr = HRESULT_FROM_WIN32(GetLastError());
        if (SUCCEEDED(hr))
            hr = pUnkWriter->QueryInterface(IID_ILITWriter,
                     reinterpret_cast<void**>(&pWriter));
        if (pUnkWriter)
            pUnkWriter->Release();
    }

    assert(SUCCEEDED(hr));

    // tell litgen where to send diagnostic messages
    if (SUCCEEDED(hr))
        hr = pWriter->SetCallback(&g_callback);

    // create the output file
    if (SUCCEEDED(hr))
        hr = pWriter->Create(pwszFilename,
                             L"c:\\temp\\",
                             L"Microsoft",
                             0);

    assert(SUCCEEDED(hr));

    // run the package file notifications
    if (SUCCEEDED(hr))
    {
        IUnknown* pUnkHost = 0;
        ILITParserHost* pHost = 0;

        hr = pWriter->GetPackageHost(TRUE, &pUnkHost);
        if (SUCCEEDED(hr))
            hr = pUnkHost->QueryInterface(IID_ILITParserHost,
                     reinterpret_cast<void**>(&pHost));
        if (SUCCEEDED(hr))
            hr = Pump(pHost, package_calls);
        if (SUCCEEDED(hr))
            hr = pHost->Finish();
        if (pHost)
            pHost->Release();
        if (pUnkHost)
            pUnkHost->Release();
    }

    assert(SUCCEEDED(hr));

    // The next sections would normally run in a loop until the hosts of each
    // type are exhausted; for simplicity, we'll unroll the loops here.

    // run the one CSS file we expect...
    if (SUCCEEDED(hr))
    {
        IUnknown* pUnkHost = 0;
        ILITCSSHost* pHost = 0;
        hr = pWriter->GetNextCSSHost(&pUnkHost);
        if (SUCCEEDED(hr))
            hr = pUnkHost->QueryInterface(IID_ILITCSSHost,
                    reinterpret_cast<void**>(&pHost));
        if (SUCCEEDED(hr))
        {
            // if we didn't already know the filename, this is where we'd
            // find out what file this host expects and open it.
            BSTR pwszFilename = 0;
            hr = pHost->GetFilename(&pwszFilename);
            if (SUCCEEDED(hr))
                std::wcout << L"Encoding CSS: " << pwszFilename << std::endl;
            SysFreeString(pwszFilename);
        }
        if (SUCCEEDED(hr))
            hr = pHost->Write(szCSS, cbCSS, 0);
        if (SUCCEEDED(hr))
            hr = pHost->Finish();
        if (pHost)
            pHost->Release();
        if (pUnkHost)
            pUnkHost->Release();
    }

    assert(SUCCEEDED(hr));

    // verify that there are no other CSS hosts available
    if (SUCCEEDED(hr))
    {
        IUnknown* pUnkDummy = 0;
        hr = pWriter->GetNextCSSHost(&pUnkDummy);
        assert(S_FALSE == hr);
        if (pUnkDummy)
            pUnkDummy->Release();
    }

    // run the content file notifications
    if (SUCCEEDED(hr))
    {
        IUnknown* pUnkHost = 0;
        ILITParserHost* pHost = 0;
        hr = pWriter->GetNextContentHost(&pUnkHost);
        if (SUCCEEDED(hr))
            hr = pUnkHost->QueryInterface(IID_ILITParserHost,
                     reinterpret_cast<void**>(&pHost));
        if (SUCCEEDED(hr))
        {
            // if we didn't already know the filename, this is where we'd
            // find out what file this host expects and open it.
            BSTR pwszFilename = 0;
            hr = pHost->GetFilename(&pwszFilename);
            if (SUCCEEDED(hr))
                std::wcout << L"Encoding content: " << pwszFilename <<
                    std::endl;
            SysFreeString(pwszFilename);
        }
        if (SUCCEEDED(hr))
            hr = Pump(pHost, pContentCalls);
        if (SUCCEEDED(hr))
            hr = pHost->Finish();
        if (pHost)
            pHost->Release();
        if (pUnkHost)
            pUnkHost->Release();
    }

    // verify that there's no other content host
    if (SUCCEEDED(hr))
    { 
        IUnknown* pUnkDummy = 0;
        hr = pWriter->GetNextContentHost(&pUnkDummy);
        assert(S_FALSE == hr);
        if (pUnkDummy)
            pUnkDummy->Release();
    }

    // we have no images here (mostly to keep the sample code simple).  If
    // we did, we'd write code like that for CSS above, without the Finish().
    if (SUCCEEDED(hr))
    {
        IUnknown* pUnkDummy = 0;
        hr = pWriter->GetNextImageHost(&pUnkDummy);
        assert(S_FALSE == hr);
        if (pUnkDummy)
            pUnkDummy->Release();
    }

    // clean up the EBWriter object
    if (SUCCEEDED(hr))
        hr = pWriter->Finish();

    if (FAILED(hr))
    {
        pWriter->Fail();
        std::wcout << L"Failed - no file generated.\n";
    }

    if (pWriter)
        pWriter->Release();
    if (hLITGen)
        FreeLibrary(hLITGen);
    CoUninitialize();
    return 0;
}

